"""
Unit Tests For Decorators
~~~~~~~~~~~~~~~~~~~~~~~~~
"""
import pytest

import argh
from argh.dto import ParserAddArgumentSpec
from argh.utils import (
    CliArgToFuncArgGuessingError,
    MixedPositionalAndOptionalArgsError,
    TooManyPositionalArgumentNames,
    naive_guess_func_arg_name,
)


def test_aliases():
    @argh.aliases("one", "two")
    def func():
        pass

    attr = getattr(func, argh.constants.ATTR_ALIASES)
    assert attr == ("one", "two")


def test_arg():
    @argh.arg("foo", help="my help", nargs="+")
    @argh.arg("--bar", default=1)
    def func():
        pass

    attrs = getattr(func, argh.constants.ATTR_ARGS)
    assert attrs == [
        ParserAddArgumentSpec(
            func_arg_name="foo",
            cli_arg_names=["foo"],
            nargs="+",
            other_add_parser_kwargs={
                "help": "my help",
            },
        ),
        ParserAddArgumentSpec(
            func_arg_name="bar",
            cli_arg_names=["--bar"],
            default_value=1,
        ),
    ]


def test_named():
    @argh.named("new-name")
    def func():
        pass

    attr = getattr(func, argh.constants.ATTR_NAME)
    assert attr == "new-name"


def test_wrap_errors():
    @argh.wrap_errors([KeyError, ValueError])
    def func():
        pass

    attr = getattr(func, argh.constants.ATTR_WRAPPED_EXCEPTIONS)
    assert attr == [KeyError, ValueError]


def test_wrap_errors_processor():
    @argh.wrap_errors(processor="STUB")
    def func():
        pass

    attr = getattr(func, argh.constants.ATTR_WRAPPED_EXCEPTIONS_PROCESSOR)
    assert attr == "STUB"


def test_naive_guess_func_arg_name() -> None:
    # none (error)
    with pytest.raises(CliArgToFuncArgGuessingError):
        argh.arg()(lambda foo: foo)

    # positional
    assert naive_guess_func_arg_name(("foo",)) == "foo"

    # positional — more than one (error)
    with pytest.raises(TooManyPositionalArgumentNames):
        argh.arg("foo", "bar")(lambda foo: foo)

    # option
    assert naive_guess_func_arg_name(("-x",)) == "x"
    assert naive_guess_func_arg_name(("--foo",)) == "foo"
    assert naive_guess_func_arg_name(("--foo", "-f")) == "foo"
    assert naive_guess_func_arg_name(("-f", "--foo")) == "foo"
    assert naive_guess_func_arg_name(("-x", "--foo", "--bar")) == "foo"

    with pytest.raises(CliArgToFuncArgGuessingError):
        naive_guess_func_arg_name(("-x", "-y"))

    # mixed (errors)
    with pytest.raises(MixedPositionalAndOptionalArgsError):
        argh.arg("foo", "--foo")(lambda foo: foo)

    with pytest.raises(MixedPositionalAndOptionalArgsError):
        argh.arg("--foo", "foo")(lambda foo: foo)

    with pytest.raises(MixedPositionalAndOptionalArgsError):
        argh.arg("-f", "foo")(lambda foo: foo)

    with pytest.raises(MixedPositionalAndOptionalArgsError):
        argh.arg("foo", "-f")(lambda foo: foo)



from typing import List, Literal, Optional, Union

import pytest

from argh.assembling import TypingHintArgSpecGuesser


@pytest.mark.parametrize("arg_type", TypingHintArgSpecGuesser.BASIC_TYPES)
def test_simple_types(arg_type):
    guess = TypingHintArgSpecGuesser.typing_hint_to_arg_spec_params

    # just the basic type
    assert guess(arg_type) == {"type": arg_type}

    # basic type or None
    assert guess(Optional[arg_type]) == {
        "type": arg_type,
        "required": False,
    }
    assert guess(Union[None, arg_type]) == {"required": False}

    # multiple basic types: the first one is used and None is looked up
    assert guess(Union[arg_type, str, None]) == {
        "type": arg_type,
        "required": False,
    }
    assert guess(Union[str, arg_type, None]) == {
        "type": str,
        "required": False,
    }


def test_list():
    guess = TypingHintArgSpecGuesser.typing_hint_to_arg_spec_params

    assert guess(list) == {"nargs": "*"}
    assert guess(Optional[list]) == {"nargs": "*", "required": False}

    assert guess(List[str]) == {"nargs": "*", "type": str}
    assert guess(List[int]) == {"nargs": "*", "type": int}
    assert guess(Optional[List[str]]) == {"nargs": "*", "type": str, "required": False}
    assert guess(Optional[List[tuple]]) == {"nargs": "*", "required": False}

    assert guess(List[list]) == {"nargs": "*"}
    assert guess(List[tuple]) == {"nargs": "*"}


def test_literal():
    guess = TypingHintArgSpecGuesser.typing_hint_to_arg_spec_params

    assert guess(Literal["a"]) == {"choices": ("a",), "type": str}
    assert guess(Literal["a", "b"]) == {"choices": ("a", "b"), "type": str}
    assert guess(Literal[1]) == {"choices": (1,), "type": int}


@pytest.mark.parametrize("arg_type", (dict, tuple))
def test_unusable_types(arg_type):
    guess = TypingHintArgSpecGuesser.typing_hint_to_arg_spec_params

    assert guess(arg_type) == {}



"""
Unit Tests For Utility Functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
from argparse import ArgumentParser, _SubParsersAction

import pytest

from argh.utils import SubparsersNotDefinedError, get_subparsers, unindent


def test_util_unindent():
    "Self-test the unindent() helper function"

    # target case
    one = """
    a
     b
      c
    """
    assert (
        unindent(one)
        == """
a
 b
  c
"""
    )

    # edge case: lack of indentation on first non-empty line
    two = """
a
  b
    c
"""

    assert unindent(two) == two

    # edge case: unexpectedly unindented in between
    three = """
    a
b
    c
    """

    assert (
        unindent(three)
        == """
a
b
c
"""
    )


def test_get_subparsers_existing() -> None:
    parser = ArgumentParser()
    parser.add_subparsers(help="hello")
    sub_parsers_action = get_subparsers(parser)
    assert isinstance(sub_parsers_action, _SubParsersAction)
    assert sub_parsers_action.help == "hello"


def test_get_subparsers_create() -> None:
    parser = ArgumentParser()
    sub_parsers_action = get_subparsers(parser, create=True)
    assert isinstance(sub_parsers_action, _SubParsersAction)


def test_get_subparsers_error() -> None:
    parser = ArgumentParser()
    with pytest.raises(SubparsersNotDefinedError):
        get_subparsers(parser)



"""
Unit Tests For the Argument DTO
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
from argh.dto import ParserAddArgumentSpec


def test_update_empty_dto() -> None:
    def stub_completer():
        ...

    dto = ParserAddArgumentSpec(
        func_arg_name="foo",
        cli_arg_names=["-f"],
    )
    other_dto = ParserAddArgumentSpec(
        func_arg_name="bar",
        cli_arg_names=["-f", "--foo"],
        is_required=True,
        default_value=123,
        nargs="+",
        other_add_parser_kwargs={"knights": "Ni!"},
        completer=stub_completer,
    )

    dto.update(other_dto)

    assert dto == ParserAddArgumentSpec(
        func_arg_name="foo",
        cli_arg_names=["-f", "--foo"],
        is_required=True,
        default_value=123,
        nargs="+",
        other_add_parser_kwargs={"knights": "Ni!"},
        completer=stub_completer,
    )


def test_update_full_dto() -> None:
    def stub_completer_one():
        ...

    def stub_completer_two():
        ...

    dto = ParserAddArgumentSpec(
        func_arg_name="foo",
        cli_arg_names=["-f"],
        nargs="?",
        is_required=True,
        default_value=123,
        other_add_parser_kwargs={"'tis but a": "scratch"},
        completer=stub_completer_one,
    )
    other_dto = ParserAddArgumentSpec(
        func_arg_name="bar",
        cli_arg_names=["-f", "--foo"],
        nargs="+",
        is_required=False,
        default_value=None,
        other_add_parser_kwargs={"knights": "Ni!"},
        completer=stub_completer_two,
    )

    dto.update(other_dto)

    assert dto == ParserAddArgumentSpec(
        func_arg_name="foo",
        cli_arg_names=["-f", "--foo"],
        is_required=False,
        default_value=None,
        nargs="+",
        other_add_parser_kwargs={"knights": "Ni!", "'tis but a": "scratch"},
        completer=stub_completer_two,
    )


class TestGetAllKwargs:
    ...


def test_make_from_kwargs_minimal() -> None:
    dto = ParserAddArgumentSpec.make_from_kwargs("foo", ["-f", "--foo"], {})

    assert dto == ParserAddArgumentSpec(
        func_arg_name="foo", cli_arg_names=["-f", "--foo"]
    )


def test_make_from_kwargs_full() -> None:
    dto = ParserAddArgumentSpec.make_from_kwargs(
        "foo",
        ["-f", "--foo"],
        {
            "action": "some action",
            "nargs": "?",
            "default": None,
            "type": str,
            "choices": [1, 2, 3],
            "required": False,
            "help": "some help",
            "metavar": "FOOOOO",
            "dest": "foo_dest",
            "some arbitrary key": "and its value",
        },
    )

    assert dto == ParserAddArgumentSpec(
        func_arg_name="foo",
        cli_arg_names=["-f", "--foo"],
        is_required=False,
        default_value=None,
        nargs="?",
        other_add_parser_kwargs={
            "action": "some action",
            "type": str,
            "choices": [1, 2, 3],
            "help": "some help",
            "metavar": "FOOOOO",
            "dest": "foo_dest",
            "some arbitrary key": "and its value",
        },
    )



#
#  Copyright © 2010—2023 Andrey Mikhaylenko and contributors
#
#  This file is part of Argh.
#
#  Argh is free software under terms of the GNU Lesser
#  General Public License version 3 (LGPLv3) as published by the Free
#  Software Foundation. See the file README.rst for copying conditions.
#
"""
Helpers
~~~~~~~
"""
import argparse
from typing import Optional, Sequence

from argh.assembling import add_commands, set_default_command
from argh.completion import autocomplete
from argh.dispatching import PARSER_FORMATTER, ArghNamespace, dispatch

__all__ = ["ArghParser"]


class ArghParser(argparse.ArgumentParser):
    """
    A subclass of :class:`argparse.ArgumentParser` with support for and a
    couple of convenience methods.

    All methods are but wrappers for stand-alone functions
    :func:`~argh.assembling.add_commands`,
    :func:`~argh.completion.autocomplete` and
    :func:`~argh.dispatching.dispatch`.

    Uses :attr:`~argh.dispatching.PARSER_FORMATTER`.
    """

    def __init__(self, *args, **kwargs) -> None:
        kwargs.setdefault("formatter_class", PARSER_FORMATTER)
        super().__init__(*args, **kwargs)

    def set_default_command(self, *args, **kwargs) -> None:
        "Wrapper for :func:`~argh.assembling.set_default_command`."
        return set_default_command(self, *args, **kwargs)

    def add_commands(self, *args, **kwargs) -> None:
        "Wrapper for :func:`~argh.assembling.add_commands`."
        return add_commands(self, *args, **kwargs)

    def autocomplete(self) -> None:
        "Wrapper for :func:`~argh.completion.autocomplete`."
        return autocomplete(self)

    def dispatch(self, *args, **kwargs) -> Optional[str]:
        "Wrapper for :func:`~argh.dispatching.dispatch`."
        return dispatch(self, *args, **kwargs)

    def parse_args(
        self,
        args: Optional[Sequence[str]] = None,
        namespace=None,
    ):
        """
        Wrapper for :meth:`argparse.ArgumentParser.parse_args`.  If `namespace`
        is not defined, :class:`argh.dispatching.ArghNamespace` is used.
        This is required for functions to be properly used as commands.
        """
        namespace = namespace or ArghNamespace()
        return super().parse_args(args, namespace)



#
#  Copyright © 2010—2023 Andrey Mikhaylenko and contributors
#
#  This file is part of Argh.
#
#  Argh is free software under terms of the GNU Lesser
#  General Public License version 3 (LGPLv3) as published by the Free
#  Software Foundation. See the file README.rst for copying conditions.
#
"""
Utilities
~~~~~~~~~
"""
import argparse
import re
from typing import Tuple


def get_subparsers(
    parser: argparse.ArgumentParser, create: bool = False
) -> argparse._SubParsersAction:
    """
    Returns the `argparse._SubParsersAction` instance for given
    :class:`argparse.ArgumentParser` instance as would have been returned by
    :meth:`argparse.ArgumentParser.add_subparsers`. The problem with the latter
    is that it only works once and raises an exception on the second attempt,
    and the public API seems to lack a method to get *existing* subparsers.

    :param create:
        If `True`, creates the subparser if it does not exist. Default if
        `False`.

    """
    # note that ArgumentParser._subparsers is *not* what is returned by
    # ArgumentParser.add_subparsers().
    if parser._subparsers:
        actions = [
            a for a in parser._actions if isinstance(a, argparse._SubParsersAction)
        ]
        return actions[0]

    if create:
        return parser.add_subparsers()

    raise SubparsersNotDefinedError()


def unindent(text: str) -> str:
    """
    Given a multi-line string, decreases indentation of all lines so that the
    first non-empty line has zero indentation and the remaining lines are
    adjusted accordingly.
    """
    match = re.match("(^|\n)( +)", text)
    if not match:
        return text
    first_line_indentation = match.group(2)
    depth = len(first_line_indentation)
    return re.sub(rf"(^|\n) {{{depth}}}", "\\1", text)


class SubparsersNotDefinedError(Exception):
    ...


def naive_guess_func_arg_name(option_strings: Tuple[str, ...]) -> str:
    def _opt_to_func_arg_name(opt: str) -> str:
        return opt.strip("-").replace("-", "_")

    if len(option_strings) == 1:
        # the only CLI arg name; adapt and use
        return _opt_to_func_arg_name(option_strings[0])

    are_args_positional = [not arg.startswith("-") for arg in option_strings]

    if any(are_args_positional) and not all(are_args_positional):
        raise MixedPositionalAndOptionalArgsError

    if all(are_args_positional):
        raise TooManyPositionalArgumentNames

    for option_string in option_strings:
        if option_string.startswith("--"):
            # prefixed long; adapt and use
            return _opt_to_func_arg_name(option_string[2:])

    raise CliArgToFuncArgGuessingError(
        f"Unable to convert opt strings {option_strings} to func arg name"
    )


class ArghError(Exception):
    ...


class CliArgToFuncArgGuessingError(ArghError):
    ...


class TooManyPositionalArgumentNames(CliArgToFuncArgGuessingError):
    ...


class MixedPositionalAndOptionalArgsError(CliArgToFuncArgGuessingError):
    ...



"""
Integration Tests
~~~~~~~~~~~~~~~~~
"""
import argparse
import re
import sys
import unittest.mock as mock
from enum import Enum

import pytest

import argh
from argh.assembling import NameMappingPolicy
from argh.exceptions import AssemblingError
from argh.utils import unindent

from .base import CmdResult as R
from .base import DebugArghParser, get_usage_string, run

if sys.version_info < (3, 10):
    HELP_OPTIONS_LABEL = "optional arguments"
else:
    HELP_OPTIONS_LABEL = "options"


def test_set_default_command_integration():
    def cmd(*, foo=1):
        return foo

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    assert run(parser, "") == R(out="1\n", err="")
    assert run(parser, "--foo 2") == R(out="2\n", err="")
    assert run(parser, "--help", exit=True) == 0


def test_set_default_command_integration_merging():
    @argh.arg("--foo", help="bar")
    def cmd(*, foo=1):
        return foo

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    assert run(parser, "") == R(out="1\n", err="")
    assert run(parser, "--foo 2") == R(out="2\n", err="")
    assert "bar" in parser.format_help()


#
# Function can be added to parser as is
#


def test_simple_function_no_args():
    def cmd():
        yield 1

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    assert run(parser, "") == R(out="1\n", err="")


def test_simple_function_positional():
    def cmd(x):
        yield x

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    assert run(parser, "", exit=True) == "the following arguments are required: x"
    assert run(parser, "foo") == R(out="foo\n", err="")


def test_simple_function_defaults():
    def cmd(*, x="foo"):
        yield x

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    assert run(parser, "") == R(out="foo\n", err="")
    assert run(parser, "bar", exit=True) == "unrecognized arguments: bar"
    assert run(parser, "--x bar") == R(out="bar\n", err="")


def test_simple_function_varargs():
    def func(*file_paths):
        # `paths` is the single positional argument with nargs="*"
        yield ", ".join(file_paths)

    parser = DebugArghParser()
    parser.set_default_command(func)

    assert run(parser, "") == R(out="\n", err="")
    assert run(parser, "foo") == R(out="foo\n", err="")
    assert run(parser, "foo bar") == R(out="foo, bar\n", err="")


def test_simple_function_kwargs():
    @argh.arg("foo")
    @argh.arg("--bar")
    def cmd(**kwargs):
        # `kwargs` contain all arguments not fitting ArgSpec.args and .varargs.
        # if ArgSpec.keywords in None, all @arg()'s will have to fit ArgSpec.args
        for k in sorted(kwargs):
            yield f"{k}: {kwargs[k]}"

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    message = "the following arguments are required: foo"
    assert run(parser, "", exit=True) == message
    assert run(parser, "hello") == R(out="bar: None\nfoo: hello\n", err="")
    assert run(parser, "--bar 123", exit=True) == message
    assert run(parser, "hello --bar 123") == R(out="bar: 123\nfoo: hello\n", err="")


def test_all_specs_in_one():
    @argh.arg("foo")
    @argh.arg("--bar")
    @argh.arg("fox")
    @argh.arg("--baz")
    def cmd(foo, *args, bar=1, **kwargs):
        yield f"foo: {foo}"
        yield f"bar: {bar}"
        yield f"*args: {args}"
        for k in sorted(kwargs):
            yield f"** {k}: {kwargs[k]}"

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    # 1) bar=1 is treated as --bar so positionals from @arg that go **kwargs
    #    will still have higher priority than bar.
    # 2) *args, a positional with nargs="*", sits between two required
    #    positionals (foo and fox), so it gets nothing.
    assert run(parser, "one two") == R(
        out="foo: one\n" "bar: 1\n" "*args: ()\n" "** baz: None\n" "** fox: two\n",
        err="",
    )

    # two required positionals (foo and fox) get an argument each and one extra
    # is left; therefore the middle one is given to *args.
    assert run(parser, "one two three") == R(
        out="foo: one\n"
        "bar: 1\n"
        "*args: ('two',)\n"
        "** baz: None\n"
        "** fox: three\n",
        err="",
    )

    # two required positionals (foo and fox) get an argument each and two extra
    # are left; both are given to *args (it's greedy).
    assert run(parser, "one two three four") == R(
        out="foo: one\n"
        "bar: 1\n"
        "*args: ('two', 'three')\n"
        "** baz: None\n"
        "** fox: four\n",
        err="",
    )


def test_arg_merged():
    """@arg merges into function signature."""

    @argh.arg("my", help="a moose once bit my sister")
    @argh.arg("-b", "--brain", help="i am made entirely of wood")
    def gumby(my, *, brain=None):
        return my, brain, "hurts"

    parser = DebugArghParser("PROG")
    parser.set_default_command(gumby)
    help_msg = parser.format_help()

    assert "a moose once bit my sister" in help_msg
    assert "i am made entirely of wood" in help_msg


def test_arg_mismatch_positional():
    """An `@arg("positional")` must match function signature."""

    @argh.arg("bogus-argument")
    def confuse_a_cat(vet, *, funny_things=123):
        return vet, funny_things

    parser = DebugArghParser("PROG")
    with pytest.raises(AssemblingError) as excinfo:
        parser.set_default_command(confuse_a_cat)

    msg = (
        "confuse_a_cat: argument bogus-argument does not fit "
        "function signature: vet, -f/--funny-things"
    )
    assert msg in str(excinfo.value)


def test_arg_mismatch_flag():
    """An `@arg("--flag")` must match function signature."""

    @argh.arg("--bogus-argument")
    def confuse_a_cat(vet, *, funny_things=123):
        return vet, funny_things

    parser = DebugArghParser("PROG")
    with pytest.raises(AssemblingError) as excinfo:
        parser.set_default_command(confuse_a_cat)

    msg = (
        "confuse_a_cat: argument --bogus-argument does not fit "
        "function signature: vet, -f/--funny-things"
    )
    assert msg in str(excinfo.value)


def test_arg_mismatch_positional_vs_flag():
    """An `@arg("arg")` must match a positional arg in function signature."""

    @argh.arg("foo")
    def func(*, foo=123):
        return foo

    parser = DebugArghParser("PROG")
    with pytest.raises(AssemblingError) as excinfo:
        parser.set_default_command(func)

    msg = (
        'func: argument "foo" declared as optional (in function signature)'
        " and positional (via decorator)"
    )
    assert msg in str(excinfo.value)


def test_arg_mismatch_flag_vs_positional():
    """An `@arg("--flag")` must match a keyword in function signature."""

    @argh.arg("--foo")
    def func(foo):
        return foo

    parser = DebugArghParser("PROG")
    with pytest.raises(AssemblingError) as excinfo:
        parser.set_default_command(func)

    msg = (
        'func: argument "foo" declared as positional (in function signature)'
        " and optional (via decorator)"
    )
    assert msg in str(excinfo.value)


class TestErrorWrapping:
    def _get_parrot(self):
        def parrot(*, dead=False):
            if dead:
                raise ValueError("this parrot is no more")
            else:
                return "beautiful plumage"

        return parrot

    def test_error_raised(self):
        parrot = self._get_parrot()

        parser = DebugArghParser()
        parser.set_default_command(parrot)

        assert run(parser, "") == R("beautiful plumage\n", "")
        with pytest.raises(ValueError) as excinfo:
            run(parser, "--dead")
        assert re.match("this parrot is no more", str(excinfo.value))

    def test_error_wrapped(self):
        parrot = self._get_parrot()
        wrapped_parrot = argh.wrap_errors([ValueError])(parrot)

        parser = DebugArghParser()
        parser.set_default_command(wrapped_parrot)

        assert run(parser, "") == R("beautiful plumage\n", "")
        assert run(parser, "--dead") == R(
            "", "ValueError: this parrot is no more\n", exit_code=1
        )

    def test_processor(self):
        parrot = self._get_parrot()
        wrapped_parrot = argh.wrap_errors([ValueError])(parrot)

        def failure(err):
            return "ERR: " + str(err) + "!"

        processed_parrot = argh.wrap_errors(processor=failure)(wrapped_parrot)

        parser = argh.ArghParser()
        parser.set_default_command(processed_parrot)

        assert run(parser, "--dead") == R(
            "", "ERR: this parrot is no more!\n", exit_code=1
        )

    def test_stderr_vs_stdout(self):
        @argh.wrap_errors([KeyError])
        def func(key):
            db = {"a": 1}
            return db[key]

        parser = argh.ArghParser()
        parser.set_default_command(func)

        assert run(parser, "a") == R(out="1\n", err="")
        assert run(parser, "b") == R(out="", err="KeyError: 'b'\n", exit_code=1)


def test_argv():
    def echo(text):
        return f"you said {text}"

    parser = DebugArghParser()
    parser.add_commands([echo])

    _argv = sys.argv

    sys.argv = sys.argv[:1] + ["echo", "hi there"]
    assert run(parser, None) == R("you said hi there\n", "")

    sys.argv = _argv


def test_commands_not_defined():
    parser = DebugArghParser()

    assert run(parser, "", {"raw_output": True}).out == parser.format_usage()
    assert run(parser, "").out == parser.format_usage()

    assert "unrecognized arguments" in run(parser, "foo", exit=True)
    assert "unrecognized arguments" in run(parser, "--foo", exit=True)


def test_command_not_chosen():
    def cmd(args):
        return 1

    parser = DebugArghParser()
    parser.add_commands([cmd])

    # returns a help message and doesn't exit
    assert "usage:" in run(parser, "").out


def test_invalid_choice():
    def cmd(args):
        return 1

    # root level command

    parser = DebugArghParser()
    parser.add_commands([cmd])

    assert "invalid choice" in run(parser, "bar", exit=True)

    # exits with an informative error
    assert run(parser, "--bar", exit=True) == "unrecognized arguments: --bar"

    # nested command

    parser = DebugArghParser()
    parser.add_commands([cmd], group_name="nest")

    assert "invalid choice" in run(parser, "nest bar", exit=True)

    # exits with an informative error
    assert run(parser, "nest --bar", exit=True) == "unrecognized arguments: --bar"


def test_unrecognized_arguments():
    def cmd():
        return 1

    # single-command parser

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    assert run(parser, "--bar", exit=True) == "unrecognized arguments: --bar"
    assert run(parser, "bar", exit=True) == "unrecognized arguments: bar"

    # multi-command parser

    parser = DebugArghParser()
    parser.add_commands([cmd])

    assert run(parser, "cmd --bar", exit=True) == "unrecognized arguments: --bar"
    assert run(parser, "cmd bar", exit=True) == "unrecognized arguments: bar"


def test_echo():
    "A simple command is resolved to a function."

    def echo(text):
        return f"you said {text}"

    parser = DebugArghParser()
    parser.add_commands([echo])

    assert run(parser, "echo foo") == R(out="you said foo\n", err="")


def test_bool_action():
    "Action `store_true`/`store_false` is inferred from default value."

    def parrot(*, dead=False):
        return "this parrot is no more" if dead else "beautiful plumage"

    parser = DebugArghParser()
    parser.add_commands([parrot])

    assert run(parser, "parrot").out == "beautiful plumage\n"
    assert run(parser, "parrot --dead").out == "this parrot is no more\n"


def test_bare_group_name():
    "A command can be resolved to a function, not a group_name."

    def hello():
        return "hello world"

    parser = DebugArghParser()
    parser.add_commands([hello], group_name="greet")

    # without arguments

    # returns a help message and doesn't exit
    assert "usage:" in run(parser, "greet").out

    # with an argument

    # exits with an informative error
    message = "unrecognized arguments: --name=world"
    assert run(parser, "greet --name=world", exit=True) == message


def test_function_under_group_name():
    "A subcommand is resolved to a function."

    def hello(*, name="world"):
        return f"Hello {name}!"

    def howdy(buddy):
        return f"Howdy {buddy}?"

    parser = DebugArghParser()
    parser.add_commands([hello, howdy], group_name="greet")

    assert run(parser, "greet hello").out == "Hello world!\n"
    assert run(parser, "greet hello --name=John").out == "Hello John!\n"
    assert run(parser, "greet hello John", exit=True) == "unrecognized arguments: John"

    # exits with an informative error
    message = "the following arguments are required: buddy"

    assert message in run(parser, "greet howdy --name=John", exit=True)
    assert run(parser, "greet howdy John").out == "Howdy John?\n"


def test_explicit_cmd_name():
    @argh.named("new-name")
    def orig_name():
        return "ok"

    parser = DebugArghParser()
    parser.add_commands([orig_name])
    assert "invalid choice" in run(parser, "orig-name", exit=True)
    assert run(parser, "new-name").out == "ok\n"


def test_aliases():
    @argh.aliases("alias2", "alias3")
    def alias1():
        return "ok"

    parser = DebugArghParser()
    parser.add_commands([alias1])

    assert run(parser, "alias1").out == "ok\n"
    assert run(parser, "alias2").out == "ok\n"
    assert run(parser, "alias3").out == "ok\n"


def test_help():
    parser = DebugArghParser()

    # assert the commands don't fail
    assert run(parser, "--help", exit=True) == 0
    assert run(parser, "greet --help", exit=True) == 0
    assert run(parser, "greet hello --help", exit=True) == 0


def test_arg_order():
    """Positional arguments are resolved in the order in which the @arg
    decorators are defined.
    """

    def cmd(foo, bar):
        return foo, bar

    parser = DebugArghParser()
    parser.set_default_command(cmd)
    assert run(parser, "foo bar").out == "foo\nbar\n"


def test_raw_output():
    "If the raw_output flag is set, no extra whitespace is added"

    def cmd(foo, bar):
        return foo, bar

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    assert run(parser, "foo bar").out == "foo\nbar\n"
    assert run(parser, "foo bar", {"raw_output": True}).out == "foobar"


def test_output_file():
    def cmd():
        return "Hello world!"

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    assert run(parser, "").out == "Hello world!\n"
    assert run(parser, "", {"output_file": None}).out == "Hello world!\n"


def test_command_error():
    def whiner_plain(*, code=1):
        raise argh.CommandError("I feel depressed.", code=code)

    def whiner_iterable():
        yield "Hello..."
        raise argh.CommandError("I feel depressed.")

    parser = DebugArghParser()
    parser.add_commands([whiner_plain, whiner_iterable])

    assert run(parser, "whiner-plain") == R(
        out="", err="CommandError: I feel depressed.\n", exit_code=1
    )
    assert run(parser, "whiner-plain --code=127") == R(
        out="", err="CommandError: I feel depressed.\n", exit_code=127
    )
    assert run(parser, "whiner-iterable") == R(
        out="Hello...\n", err="CommandError: I feel depressed.\n", exit_code=1
    )


@pytest.mark.parametrize(
    "argparse_namespace_class", [argparse.Namespace, argh.dispatching.ArghNamespace]
)
def test_get_function_from_namespace_obj(argparse_namespace_class):
    argparse_namespace = argparse_namespace_class()

    def func():
        pass

    retval = argh.dispatching._get_function_from_namespace_obj(argparse_namespace)
    assert retval is None

    setattr(argparse_namespace, argh.constants.DEST_FUNCTION, "")

    retval = argh.dispatching._get_function_from_namespace_obj(argparse_namespace)
    assert retval is None

    setattr(argparse_namespace, argh.constants.DEST_FUNCTION, func)

    retval = argh.dispatching._get_function_from_namespace_obj(argparse_namespace)
    assert retval == func


def test_normalized_keys():
    """Underscores in function args are converted to dashes and back."""

    def cmd(a_b):
        return a_b

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    assert run(parser, "hello").out == "hello\n"


@mock.patch("argh.assembling.COMPLETION_ENABLED", True)
def test_custom_argument_completer():
    "Issue #33: Enable custom per-argument shell completion"

    @argh.arg("foo", completer="STUB")
    def func(foo):
        pass

    parser = argh.ArghParser()
    parser.set_default_command(func)

    assert parser._actions[-1].completer == "STUB"


def test_class_members():
    "Issue #34: class members as commands"

    class Controller:
        var = 123

        def instance_meth(self, value):
            return value, self.var

        @classmethod
        def class_meth(cls, value):
            return value, cls.var

        @staticmethod
        def static_meth(value):
            return value, "w00t?"

        @staticmethod
        def static_meth2(value):
            return value, "huh!"

    controller = Controller()

    parser = DebugArghParser()
    parser.add_commands(
        [
            controller.instance_meth,
            controller.class_meth,
            controller.static_meth,
            Controller.static_meth2,
        ]
    )

    assert run(parser, "instance-meth foo").out == "foo\n123\n"
    assert run(parser, "class-meth foo").out == "foo\n123\n"
    assert run(parser, "static-meth foo").out == "foo\nw00t?\n"
    assert run(parser, "static-meth2 foo").out == "foo\nhuh!\n"


def test_kwonlyargs__policy_legacy():
    "Correct dispatch in presence of keyword-only arguments"

    def cmd(*args, foo="1", bar, baz="3", **kwargs):
        return f"foo='{foo}' bar='{bar}' baz='{baz}' args={args} kwargs={kwargs}"

    parser = DebugArghParser(prog="pytest")
    parser.set_default_command(
        cmd, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT
    )

    expected_usage = "usage: pytest [-h] [-f FOO] [--baz BAZ] [args ...] bar\n"
    if sys.version_info < (3, 9):
        # https://github.com/python/cpython/issues/82619
        expected_usage = (
            "usage: pytest [-h] [-f FOO] [--baz BAZ] [args [args ...]] bar\n"
        )
    assert parser.format_usage() == expected_usage

    assert (
        run(parser, "--baz=baz! one  two").out
        == "foo='1' bar='two' baz='baz!' args=('one',) kwargs={}\n"
    )
    assert (
        run(parser, "test --foo=do").out
        == "foo='do' bar='test' baz='3' args=() kwargs={}\n"
    )


def test_kwonlyargs__policy_modern():
    "Correct dispatch in presence of keyword-only arguments"

    def cmd(*args, foo="1", bar, baz="3", **kwargs):
        return f"foo='{foo}' bar='{bar}' baz='{baz}' args={args} kwargs={kwargs}"

    parser = DebugArghParser(prog="pytest")
    parser.set_default_command(
        cmd, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_KWONLY
    )

    expected_usage = "usage: pytest [-h] [-f FOO] --bar BAR [--baz BAZ] [args ...]\n"
    if sys.version_info < (3, 9):
        # https://github.com/python/cpython/issues/82619
        expected_usage = (
            "usage: pytest [-h] [-f FOO] --bar BAR [--baz BAZ] [args [args ...]]\n"
        )
    assert parser.format_usage() == expected_usage

    assert (
        run(parser, "--baz=baz! one  two --bar=bar!").out
        == "foo='1' bar='bar!' baz='baz!' args=('one', 'two') kwargs={}\n"
    )
    message = "the following arguments are required: --bar"
    assert run(parser, "test --foo=do", exit=True) == message


def test_default_arg_values_in_help():
    "Argument defaults should appear in the help message implicitly"

    @argh.arg("name", default="Basil")
    @argh.arg("--task", default="hang the Moose")
    @argh.arg("--note", help="why is it a remarkable animal?")
    def remind(
        name,
        *,
        task=None,
        reason="there are creatures living in it",
        note="it can speak English",
    ):
        return "Oh what is it now, can't you leave me in peace..."

    parser = DebugArghParser()
    parser.set_default_command(remind)

    help_normalised = re.sub(r"\s+", " ", parser.format_help())

    assert "name 'Basil'" in help_normalised
    assert "-t TASK, --task TASK 'hang the Moose'" in help_normalised
    assert (
        "-r REASON, --reason REASON 'there are creatures living in it'"
        in help_normalised
    )

    # explicit help message is not obscured by the implicit one
    # but is still present
    assert (
        "-n NOTE, --note NOTE why is it a remarkable animal? "
        "(default: 'it can speak English')"
    ) in help_normalised


def test_default_arg_values_in_help__regression():
    "Empty string as default value → empty help string → broken argparse"

    def foo(*, bar=""):
        return bar

    parser = DebugArghParser()
    parser.set_default_command(foo)

    # doesn't break
    parser.format_help()

    # now check details
    assert "-b BAR, --bar BAR  ''" in parser.format_help()
    # note the empty str repr ^^^


def test_help_formatting_is_preserved():
    "Formatting of docstrings should not be messed up in help messages"

    def func():
        """
        Sample function.

        Parameters:
            foo: float
                An example argument.
            bar: bool
                Another argument.
        """
        return "hello"

    parser = DebugArghParser()
    parser.set_default_command(func)

    assert unindent(func.__doc__) in parser.format_help()


def test_prog(capsys: pytest.CaptureFixture[str]):
    "Program name propagates from sys.argv[0]"

    def cmd(*, foo=1):
        return foo

    parser = DebugArghParser()
    parser.add_commands([cmd])

    usage = get_usage_string()

    assert run(parser, "-h", exit=True) == 0
    captured = capsys.readouterr()
    assert captured.out.startswith(usage)


def test_unknown_args():
    def cmd(*, foo=1):
        return foo

    parser = DebugArghParser()
    parser.set_default_command(cmd)

    get_usage_string("[-f FOO]")

    assert run(parser, "--foo 1") == R(out="1\n", err="")
    assert run(parser, "--bar 1", exit=True) == "unrecognized arguments: --bar 1"
    assert run(parser, "--bar 1", exit=False, kwargs={"skip_unknown_args": True}) == R(
        out="1\n", err=""
    )


def test_add_commands_unknown_name_mapping_policy():
    def func(foo):
        ...

    parser = argh.ArghParser(prog="myapp")

    class UnsuitablePolicyContainer(Enum):
        FOO = "Ni!!!"

    with pytest.raises(
        NotImplementedError,
        match="Unknown name mapping policy UnsuitablePolicyContainer.FOO",
    ):
        parser.add_commands([func], name_mapping_policy=UnsuitablePolicyContainer.FOO)


def test_add_commands_no_overrides1(capsys: pytest.CaptureFixture[str]):
    def first_func(*, foo=123):
        """Owl stretching time"""

    def second_func():
        pass

    parser = argh.ArghParser(prog="myapp")
    parser.add_commands(
        [first_func, second_func],
    )

    run(parser, "--help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp [-h] {{first-func,second-func}} ...

        positional arguments:
          {{first-func,second-func}}
            first-func          Owl stretching time
            second-func

        {HELP_OPTIONS_LABEL}:
          -h, --help            show this help message and exit
        """
        )[1:]
    )


def test_add_commands_no_overrides2(capsys: pytest.CaptureFixture[str]):
    def first_func(*, foo=123):
        """Owl stretching time"""

    def second_func():
        pass

    parser = argh.ArghParser(prog="myapp")
    parser.add_commands([first_func, second_func])

    run(parser, "first-func --help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp first-func [-h] [-f FOO]

        Owl stretching time

        {HELP_OPTIONS_LABEL}:
          -h, --help         show this help message and exit
          -f FOO, --foo FOO  123
        """
        )[1:]
    )


def test_add_commands_group_overrides1(capsys: pytest.CaptureFixture[str]):
    """
    When `group_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(*, foo=123):
        """Owl stretching time"""
        return foo

    def second_func():
        pass

    parser = argh.ArghParser(prog="myapp")
    parser.add_commands(
        [first_func, second_func],
        group_name="my-group",
        group_kwargs={
            "help": "group help override",
            "description": "group description override",
        },
    )

    run(parser, "--help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp [-h] {{my-group}} ...

        positional arguments:
          {{my-group}}
            my-group

        {HELP_OPTIONS_LABEL}:
          -h, --help  show this help message and exit
        """
        )[1:]
    )


def test_add_commands_group_overrides2(capsys: pytest.CaptureFixture[str]):
    """
    When `group_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(*, foo=123):
        """Owl stretching time"""
        return foo

    def second_func():
        pass

    parser = argh.ArghParser(prog="myapp")
    parser.add_commands(
        [first_func, second_func],
        group_name="my-group",
        group_kwargs={
            "help": "group help override",
            "description": "group description override",
        },
    )

    run(parser, "my-group --help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp my-group [-h] {{first-func,second-func}} ...

        {HELP_OPTIONS_LABEL}:
          -h, --help            show this help message and exit

        subcommands:
          group description override

          {{first-func,second-func}}
                                group help override
            first-func          Owl stretching time
            second-func
        """
        )[1:]
    )


def test_add_commands_group_overrides3(capsys: pytest.CaptureFixture[str]):
    """
    When `group_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(*, foo=123):
        """Owl stretching time"""
        return foo

    def second_func():
        pass

    parser = argh.ArghParser(prog="myapp")
    parser.add_commands(
        [first_func, second_func],
        group_name="my-group",
        group_kwargs={
            "help": "group help override",
            "description": "group description override",
        },
    )

    run(parser, "my-group first-func --help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp my-group first-func [-h] [-f FOO]

        Owl stretching time

        {HELP_OPTIONS_LABEL}:
          -h, --help         show this help message and exit
          -f FOO, --foo FOO  123
        """
        )[1:]
    )


def test_add_commands_func_overrides1(capsys: pytest.CaptureFixture[str]):
    """
    When `func_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(*, foo=123):
        """Owl stretching time"""
        pass

    def second_func():
        pass

    parser = argh.ArghParser(prog="myapp")
    parser.add_commands(
        [first_func, second_func],
        func_kwargs={
            "help": "func help override",
            "description": "func description override",
        },
    )

    run(parser, "--help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp [-h] {{first-func,second-func}} ...

        positional arguments:
          {{first-func,second-func}}
            first-func          func help override
            second-func         func help override

        {HELP_OPTIONS_LABEL}:
          -h, --help            show this help message and exit
        """
        )[1:]
    )


def test_add_commands_func_overrides2(capsys: pytest.CaptureFixture[str]):
    """
    When `func_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(*, foo=123):
        """Owl stretching time"""
        pass

    def second_func():
        pass

    parser = argh.ArghParser(prog="myapp")
    parser.add_commands(
        [first_func, second_func],
        func_kwargs={
            "help": "func help override",
            "description": "func description override",
        },
    )

    run(parser, "first-func --help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp first-func [-h] [-f FOO]

        func description override

        {HELP_OPTIONS_LABEL}:
          -h, --help         show this help message and exit
          -f FOO, --foo FOO  123
        """
        )[1:]
    )


def test_action_count__only_arg_decorator():
    @argh.arg("-v", "--verbose", action="count", default=0)
    def func(**kwargs):
        verbosity = kwargs.get("verbose")
        return f"verbosity: {verbosity}"

    parser = DebugArghParser()
    parser.set_default_command(func)

    assert run(parser, "").out == "verbosity: 0\n"
    assert run(parser, "-v").out == "verbosity: 1\n"
    assert run(parser, "-vvvv").out == "verbosity: 4\n"


def test_action_count__mixed():
    @argh.arg("-v", "--verbose", action="count")
    def func(*, verbose=0):
        return f"verbosity: {verbose}"

    parser = DebugArghParser()
    parser.set_default_command(func)

    assert run(parser, "").out == "verbosity: 0\n"
    assert run(parser, "-v").out == "verbosity: 1\n"
    assert run(parser, "-vvvv").out == "verbosity: 4\n"



"""
Unit Tests For Assembling Phase
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
import argparse
from typing import Literal, Optional
from unittest.mock import MagicMock, call, patch

import pytest

import argh
from argh.assembling import AssemblingError, NameMappingPolicy
from argh.dto import ParserAddArgumentSpec


def test_guess_type_from_choices():
    given = ParserAddArgumentSpec(
        "foo", ["foo"], other_add_parser_kwargs={"choices": [1, 2]}
    )
    guessed = {"type": int}
    assert guessed == argh.assembling.guess_extra_parser_add_argument_spec_kwargs(given)

    # do not override a guessable param if already explicitly defined
    given = ParserAddArgumentSpec(
        "foo",
        ["foo"],
        other_add_parser_kwargs={
            "option_strings": ["foo"],
            "choices": [1, 2],
            "type": "NO_MATTER_WHAT",
        },
    )
    guessed = {}
    assert guessed == argh.assembling.guess_extra_parser_add_argument_spec_kwargs(given)


def test_guess_type_from_default():
    given = ParserAddArgumentSpec("foo", ["foo"], default_value=1)
    guessed = {"type": int}
    assert guessed == argh.assembling.guess_extra_parser_add_argument_spec_kwargs(given)

    # do not override a guessable param if already explicitly defined
    given = ParserAddArgumentSpec(
        "foo",
        ["foo"],
        default_value=1,
        other_add_parser_kwargs={
            "type": "NO_MATTER_WHAT",
        },
    )
    guessed = {}
    assert guessed == argh.assembling.guess_extra_parser_add_argument_spec_kwargs(given)


def test_guess_action_from_default():
    # positional, default True → ignore
    given = ParserAddArgumentSpec("foo", ["foo"], default_value=False)
    assert {} == argh.assembling.guess_extra_parser_add_argument_spec_kwargs(given)

    # named, default True → store_false
    given = ParserAddArgumentSpec("foo", ["--foo"], default_value=False)
    guessed = {"action": "store_true"}
    assert guessed == argh.assembling.guess_extra_parser_add_argument_spec_kwargs(given)

    # positional, default False → ignore
    given = ParserAddArgumentSpec("foo", ["foo"], default_value=True)
    assert {} == argh.assembling.guess_extra_parser_add_argument_spec_kwargs(given)

    # named, True → store_false
    given = ParserAddArgumentSpec("foo", ["--foo"], default_value=True)
    guessed = {"action": "store_false"}
    assert guessed == argh.assembling.guess_extra_parser_add_argument_spec_kwargs(given)

    # do not override a guessable param if already explicitly defined
    given = ParserAddArgumentSpec(
        "foo",
        ["foo"],
        default_value=False,
        other_add_parser_kwargs={
            "action": "NO_MATTER_WHAT",
        },
    )
    guessed = {}
    assert guessed == argh.assembling.guess_extra_parser_add_argument_spec_kwargs(given)


def test_positional_with_default_int():
    def func(pos_int_default=123):
        ...

    parser = argh.ArghParser(prog="test")
    parser.set_default_command(
        func, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_KWONLY
    )
    assert parser.format_usage() == "usage: test [-h] [pos-int-default]\n"
    assert "pos-int-default  123" in parser.format_help()


def test_positional_with_default_bool():
    def func(pos_bool_default=False):
        ...

    parser = argh.ArghParser(prog="test")
    parser.set_default_command(
        func, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_KWONLY
    )
    assert parser.format_usage() == "usage: test [-h] [pos-bool-default]\n"
    assert "pos-bool-default  False" in parser.format_help()


def test_set_default_command():
    def func(**kwargs):
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        [
            ParserAddArgumentSpec(
                func_arg_name="foo",
                cli_arg_names=("foo",),
                nargs=argparse.ONE_OR_MORE,
                other_add_parser_kwargs={"choices": [1, 2], "help": "me"},
            ),
            ParserAddArgumentSpec(
                func_arg_name="bar", cli_arg_names=("-b", "--bar"), default_value=False
            ),
        ],
    )

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()
    parser.set_defaults = MagicMock()

    argh.set_default_command(parser, func)

    assert parser.add_argument.mock_calls == [
        call("foo", nargs=argparse.ONE_OR_MORE, choices=[1, 2], help="me", type=int),
        call(
            "-b",
            "--bar",
            default=False,
            action="store_true",
            help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE,
        ),
    ]
    assert parser.set_defaults.mock_calls == [call(function=func)]


def test_set_default_command__parser_error():
    def func(foo: str) -> str:
        return foo

    parser_mock = MagicMock(spec=argparse.ArgumentParser)
    parser_mock.add_help = False
    parser_mock.add_argument.side_effect = argparse.ArgumentError(
        None, "my hat's on fire!"
    )

    with pytest.raises(argh.AssemblingError):
        argh.set_default_command(parser_mock, func)


def test_set_default_command__no_func_args():
    # TODO: document in changelog
    # XXX breaking change in v0.30!
    #     Old behaviour: @arg declarations would be passed to add_argument().
    #                    (how the hell would it look like though?)
    #     New behaviour: @arg declarations are ignored because there's no func
    #                    arg to map them onto.
    def func():
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        [ParserAddArgumentSpec(func_arg_name="x", cli_arg_names=("-x",))],
    )

    p = argh.ArghParser()

    with pytest.raises(argh.AssemblingError) as excinfo:
        p.set_default_command(func)
    msg = (
        "func: cannot extend argument declarations for an endpoint "
        "function that takes no arguments."
    )
    assert msg in str(excinfo.value)


def test_set_default_command__varargs_vs_positional():
    def func(*args):
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        [ParserAddArgumentSpec(func_arg_name="x", cli_arg_names=("x",))],
    )

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()
    parser.set_defaults = MagicMock()

    with pytest.raises(
        AssemblingError, match="func: argument x does not fit function signature: args"
    ):
        parser.set_default_command(func)


def test_set_default_command__varargs_vs_optional():
    def func(*args):
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        [ParserAddArgumentSpec(func_arg_name="x", cli_arg_names=("-x",))],
    )

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()
    parser.set_defaults = MagicMock()

    with pytest.raises(
        AssemblingError, match="func: argument -x does not fit function signature: args"
    ):
        parser.set_default_command(func)


def test_set_default_command__varkwargs_vs_positional():
    def func(**kwargs):
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        [ParserAddArgumentSpec(func_arg_name="x", cli_arg_names=("x",))],
    )

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()
    parser.set_defaults = MagicMock()

    parser.set_default_command(func)
    assert parser.add_argument.mock_calls == [call("x", help="%(default)s")]
    assert parser.set_defaults.mock_calls == [call(function=func)]


def test_set_default_command__varkwargs_vs_optional():
    def func(**kwargs):
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        [ParserAddArgumentSpec(func_arg_name="x", cli_arg_names=("-x",))],
    )

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()
    parser.set_defaults = MagicMock()

    parser.set_default_command(func)
    assert parser.add_argument.mock_calls == [call("-x", help="%(default)s")]
    assert parser.set_defaults.mock_calls == [call(function=func)]


def test_set_default_command__declared_vs_signature__names_mismatch():
    def func(bar):
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        (
            ParserAddArgumentSpec(
                func_arg_name="x",
                cli_arg_names=("foo",),
                nargs=argparse.ONE_OR_MORE,
                other_add_parser_kwargs={"choices": [1, 2], "help": "me"},
            ),
        ),
    )

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()
    parser.set_defaults = MagicMock()

    with pytest.raises(
        AssemblingError, match="func: argument foo does not fit function signature: bar"
    ):
        argh.set_default_command(parser, func)


def test_set_default_command__declared_vs_signature__same_name_pos_vs_opt():
    def func(foo):
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        (ParserAddArgumentSpec(func_arg_name="foo", cli_arg_names=("--foo",)),),
    )

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()
    parser.set_defaults = MagicMock()

    import re

    with pytest.raises(
        AssemblingError,
        match=re.escape(
            'func: argument "foo" declared as positional (in function signature) and optional (via decorator)'
        ),
    ):
        argh.set_default_command(parser, func)


@pytest.fixture()
def big_command_with_everything():
    # TODO: split into small tests where we'd check each combo and make sure
    # they interact as expected (e.g. pos opt arg gets the short form even if
    # there's a pos req arg, etc.)
    #
    # an arg with a unique first letter per arg type + every combination
    # of conflicting first letters per every two arg types:
    # - positional required (i.e. without a default value)
    # - positional optional (i.e. with a default value)
    # - named-only required (i.e. kwonly without a default value)
    # - named-only optional (i.e. kwonly with a default value)
    def func(
        alpha_pos_req,
        beta_pos_req,
        alpha_pos_opt="alpha",
        beta_pos_opt_one="beta one",
        beta_pos_opt_two="beta two",
        gamma_pos_opt="gamma named",
        delta_pos_opt="delta named",
        theta_pos_opt="theta named",
        *args,
        gamma_kwonly_opt="gamma kwonly",
        delta_kwonly_req,
        epsilon_kwonly_req_one,
        epsilon_kwonly_req_two,
        zeta_kwonly_opt="zeta kwonly",
        **kwargs,
    ):
        return (
            alpha_pos_req,
            beta_pos_req,
            alpha_pos_opt,
            beta_pos_opt_one,
            beta_pos_opt_two,
            gamma_pos_opt,
            delta_pos_opt,
            theta_pos_opt,
            gamma_kwonly_opt,
            delta_kwonly_req,
            epsilon_kwonly_req_one,
            epsilon_kwonly_req_two,
            zeta_kwonly_opt,
            args,
            kwargs,
        )

    yield func


def test_set_default_command_infer_cli_arg_names_from_func_signature__policy_legacy(
    big_command_with_everything,
):
    name_mapping_policy = NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()
    parser.set_defaults = MagicMock()

    argh.set_default_command(
        parser, big_command_with_everything, name_mapping_policy=name_mapping_policy
    )

    help_tmpl = argh.constants.DEFAULT_ARGUMENT_TEMPLATE
    assert parser.add_argument.mock_calls == [
        call("alpha-pos-req", help="%(default)s"),
        call("beta-pos-req", help="%(default)s"),
        call("-a", "--alpha-pos-opt", default="alpha", type=str, help=help_tmpl),
        call("--beta-pos-opt-one", default="beta one", type=str, help=help_tmpl),
        call("--beta-pos-opt-two", default="beta two", type=str, help=help_tmpl),
        call("--gamma-pos-opt", default="gamma named", type=str, help=help_tmpl),
        call("--delta-pos-opt", default="delta named", type=str, help=help_tmpl),
        call("-t", "--theta-pos-opt", default="theta named", type=str, help=help_tmpl),
        call("args", nargs=argparse.ZERO_OR_MORE, help=help_tmpl),
        call("--gamma-kwonly-opt", default="gamma kwonly", type=str, help=help_tmpl),
        call("delta-kwonly-req", help=help_tmpl),
        call("epsilon-kwonly-req-one", help=help_tmpl),
        call("epsilon-kwonly-req-two", help=help_tmpl),
        call(
            "-z", "--zeta-kwonly-opt", default="zeta kwonly", type=str, help=help_tmpl
        ),
    ]
    assert parser.set_defaults.mock_calls == [
        call(function=big_command_with_everything)
    ]


def test_set_default_command_infer_cli_arg_names_from_func_signature__policy_modern(
    big_command_with_everything,
):
    name_mapping_policy = NameMappingPolicy.BY_NAME_IF_KWONLY

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()
    parser.set_defaults = MagicMock()

    argh.set_default_command(
        parser, big_command_with_everything, name_mapping_policy=name_mapping_policy
    )

    help_tmpl = argh.constants.DEFAULT_ARGUMENT_TEMPLATE
    assert parser.add_argument.mock_calls == [
        call("alpha-pos-req", help="%(default)s"),
        call("beta-pos-req", help="%(default)s"),
        call(
            "alpha-pos-opt",
            default="alpha",
            nargs=argparse.OPTIONAL,
            type=str,
            help=help_tmpl,
        ),
        call(
            "beta-pos-opt-one",
            default="beta one",
            nargs=argparse.OPTIONAL,
            type=str,
            help=help_tmpl,
        ),
        call(
            "beta-pos-opt-two",
            default="beta two",
            nargs=argparse.OPTIONAL,
            type=str,
            help=help_tmpl,
        ),
        call(
            "gamma-pos-opt",
            default="gamma named",
            nargs=argparse.OPTIONAL,
            type=str,
            help=help_tmpl,
        ),
        call(
            "delta-pos-opt",
            default="delta named",
            nargs=argparse.OPTIONAL,
            type=str,
            help=help_tmpl,
        ),
        call(
            "theta-pos-opt",
            default="theta named",
            nargs=argparse.OPTIONAL,
            type=str,
            help=help_tmpl,
        ),
        call("args", nargs=argparse.ZERO_OR_MORE, help=help_tmpl),
        call("--gamma-kwonly-opt", default="gamma kwonly", type=str, help=help_tmpl),
        call("--delta-kwonly-req", required=True, help=help_tmpl),
        call("--epsilon-kwonly-req-one", required=True, help=help_tmpl),
        call("--epsilon-kwonly-req-two", required=True, help=help_tmpl),
        call(
            "-z", "--zeta-kwonly-opt", default="zeta kwonly", type=str, help=help_tmpl
        ),
    ]
    assert parser.set_defaults.mock_calls == [
        call(function=big_command_with_everything)
    ]


def test_set_default_command_docstring():
    def func():
        "docstring"
        pass

    parser = argh.ArghParser()

    argh.set_default_command(parser, func)

    assert parser.description == "docstring"


def test_set_default_command__varkwargs_sharing_prefix():
    def func(*, alpha: str = "Alpha", aleph: str = "Aleph"):
        ...

    parser = argh.ArghParser()
    parser.add_argument = MagicMock()

    argh.set_default_command(parser, func)

    assert parser.add_argument.mock_calls == [
        call("--alpha", default="Alpha", type=str, help="%(default)s"),
        call("--aleph", default="Aleph", type=str, help="%(default)s"),
    ]


def test_add_subparsers_when_default_command_exists():
    def one():
        return 1

    def two():
        return 2

    def three():
        return 3

    p = argh.ArghParser()
    p.set_default_command(one)
    p.add_commands([two, three])

    ns_default = p.parse_args([])

    ns_explicit_two = p.parse_args(["two"])
    ns_explicit_three = p.parse_args(["three"])

    assert ns_default.get_function() == one
    assert ns_explicit_two.get_function() == two
    assert ns_explicit_three.get_function() == three


def test_set_default_command_when_subparsers_exist():
    def one():
        return 1

    def two():
        return 2

    def three():
        return 3

    p = argh.ArghParser()
    p.add_commands([one, two])
    p.set_default_command(three)

    ns_default = p.parse_args([])
    ns_explicit_one = p.parse_args(["one"])
    ns_explicit_two = p.parse_args(["two"])

    assert ns_explicit_one.get_function() == one
    assert ns_explicit_two.get_function() == two
    assert ns_default.get_function() == three


def test_set_add_commands_twice():
    def one():
        return 1

    def two():
        return 2

    p = argh.ArghParser()
    p.add_commands([one])
    p.add_commands([two])

    ns_explicit_one = p.parse_args(["one"])
    ns_explicit_two = p.parse_args(["two"])

    assert ns_explicit_one.get_function() == one
    assert ns_explicit_two.get_function() == two


def test_add_command_with_group_kwargs_but_no_group_name():
    def one():
        return 1

    p = argh.ArghParser()
    err_msg = "`group_kwargs` only makes sense with `group_name`"
    with pytest.raises(ValueError, match=err_msg):
        p.add_commands([one], group_kwargs={"help": "foo"})


def test_set_default_command_varargs():
    def func(*file_paths):
        yield ", ".join(file_paths)

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()

    argh.set_default_command(parser, func)

    assert parser.add_argument.mock_calls == [
        call(
            "file-paths",
            nargs=argparse.ZERO_OR_MORE,
            help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE,
        ),
    ]


def test_set_default_command_kwargs():
    @argh.arg("foo")
    @argh.arg("--bar")
    def func(x, **kwargs):
        pass

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()

    argh.set_default_command(parser, func)

    assert parser.add_argument.mock_calls == [
        call("x", help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE),
        call("foo", help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE),
        call("--bar", help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE),
    ]


def test_kwonlyargs__policy_legacy():
    "Correctly processing required and optional keyword-only arguments"

    def cmd(foo_pos, bar_pos, *args, foo_kwonly="foo_kwonly", bar_kwonly):
        return (foo_pos, bar_pos, args, foo_kwonly, bar_kwonly)

    parser = argh.ArghParser()
    parser.add_argument = MagicMock()
    parser.set_default_command(
        cmd, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT
    )
    help_tmpl = argh.constants.DEFAULT_ARGUMENT_TEMPLATE
    assert parser.add_argument.mock_calls == [
        call("foo-pos", help=help_tmpl),
        call("bar-pos", help=help_tmpl),
        call("args", nargs=argparse.ZERO_OR_MORE, help=help_tmpl),
        call("-f", "--foo-kwonly", default="foo_kwonly", type=str, help=help_tmpl),
        call("bar-kwonly", help=help_tmpl),
    ]


def test_kwonlyargs__policy_modern():
    "Correctly processing required and optional keyword-only arguments"

    def cmd(foo_pos, bar_pos, *args, foo_kwonly="foo_kwonly", bar_kwonly):
        return (foo_pos, bar_pos, args, foo_kwonly, bar_kwonly)

    parser = argh.ArghParser()
    parser.add_argument = MagicMock()
    parser.set_default_command(
        cmd, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_KWONLY
    )
    help_tmpl = argh.constants.DEFAULT_ARGUMENT_TEMPLATE
    assert parser.add_argument.mock_calls == [
        call("foo-pos", help=help_tmpl),
        call("bar-pos", help=help_tmpl),
        call("args", nargs=argparse.ZERO_OR_MORE, help=help_tmpl),
        call("-f", "--foo-kwonly", default="foo_kwonly", type=str, help=help_tmpl),
        call("-b", "--bar-kwonly", required=True, help=help_tmpl),
    ]


@patch("argh.assembling.COMPLETION_ENABLED", True)
def test_custom_argument_completer():
    "Issue #33: Enable custom per-argument shell completion"

    def func(foo):
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        [
            ParserAddArgumentSpec(
                func_arg_name="foo", cli_arg_names=("foo",), completer="STUB"
            )
        ],
    )

    p = argh.ArghParser()
    p.set_default_command(func)

    assert p._actions[-1].completer == "STUB"


@patch("argh.assembling.COMPLETION_ENABLED", False)
def test_custom_argument_completer_no_backend():
    "If completion backend is not available, nothing breaks"

    def func(foo):
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        [
            ParserAddArgumentSpec(
                func_arg_name="foo", cli_arg_names=("foo",), completer="STUB"
            )
        ],
    )

    p = argh.ArghParser()
    p.set_default_command(func)

    assert not hasattr(p._actions[-1], "completer")


@patch("argh.assembling.add_commands")
def test_add_subcommands(mock_add_commands):
    mock_parser = MagicMock()

    def get_items():
        pass

    argh.add_subcommands(
        mock_parser,
        "db",
        [get_items],
        title="database commands",
        help="CRUD for our silly database",
    )

    mock_add_commands.assert_called_with(
        mock_parser,
        [get_items],
        group_name="db",
        group_kwargs={
            "title": "database commands",
            "help": "CRUD for our silly database",
        },
    )


@patch("argh.helpers.autocomplete")
def test_arghparser_autocomplete_method(mock_autocomplete):
    p = argh.ArghParser()
    p.autocomplete()

    mock_autocomplete.assert_called()


def test_is_positional():
    with pytest.raises(ValueError, match="Expected at least one"):
        argh.assembling._is_positional([])
    with pytest.raises(ValueError, match="Expected at least one"):
        argh.assembling._is_positional([""])
    assert argh.assembling._is_positional(["f"]) is True
    assert argh.assembling._is_positional(["foo"]) is True
    assert argh.assembling._is_positional(["--foo"]) is False
    assert argh.assembling._is_positional(["-f"]) is False
    assert argh.assembling._is_positional(["-f", "--foo"]) is False

    # this spec is invalid but validation is out of scope of the function
    # as it only checks if the first argument has the leading dash
    assert argh.assembling._is_positional(["-f", "foo"]) is False


def test_typing_hints_only_used_when_arg_deco_not_used():
    @argh.arg("foo", type=int)
    def func_decorated(foo: Optional[float]):
        ...

    def func_undecorated(bar: Optional[float]):
        ...

    parser = argparse.ArgumentParser()
    parser.add_argument = MagicMock()
    argh.set_default_command(parser, func_decorated)
    assert parser.add_argument.mock_calls == [
        call("foo", type=int, help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE),
    ]

    parser = argparse.ArgumentParser()
    parser.add_argument = MagicMock()
    argh.set_default_command(parser, func_undecorated)
    assert parser.add_argument.mock_calls == [
        call(
            "bar",
            nargs="?",
            type=float,
            help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE,
        ),
    ]


def test_typing_hints_overview():
    def func(
        alpha,
        beta: str,
        gamma: Optional[int] = None,
        *,
        delta: float = 1.5,
        epsilon: Optional[int] = 42,
        zeta: bool = False,
    ) -> str:
        return f"alpha={alpha}, beta={beta}, gamma={gamma}, delta={delta}, epsilon={epsilon}, zeta={zeta}"

    parser = argparse.ArgumentParser()
    parser.add_argument = MagicMock()
    argh.set_default_command(
        parser, func, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_KWONLY
    )
    _extra_kw = {"help": argh.constants.DEFAULT_ARGUMENT_TEMPLATE}
    assert parser.add_argument.mock_calls == [
        call("alpha", **_extra_kw),
        call("beta", type=str, **_extra_kw),
        call("gamma", default=None, nargs="?", type=int, **_extra_kw),
        call("-d", "--delta", type=float, default=1.5, **_extra_kw),
        call("-e", "--epsilon", type=int, default=42, required=False, **_extra_kw),
        call("-z", "--zeta", default=False, action="store_true", **_extra_kw),
    ]


def test_typing_hints_str__policy_by_name_if_has_default():
    def func(alpha: str, beta: str = "N/A", *, gamma: str, delta: str = "N/A") -> str:
        return f"alpha={alpha}, beta={beta}, gamma={gamma}, delta={delta}"

    parser = argparse.ArgumentParser()
    parser.add_argument = MagicMock()
    argh.set_default_command(
        parser, func, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT
    )
    _extra_kw = {"help": argh.constants.DEFAULT_ARGUMENT_TEMPLATE}
    assert parser.add_argument.mock_calls == [
        call("alpha", type=str, **_extra_kw),
        call("-b", "--beta", default="N/A", type=str, **_extra_kw),
        call("gamma", type=str, **_extra_kw),
        call("-d", "--delta", default="N/A", type=str, **_extra_kw),
    ]


def test_typing_hints_str__policy_by_name_if_kwonly():
    def func(alpha: str, beta: str = "N/A", *, gamma: str, delta: str = "N/A") -> str:
        return f"alpha={alpha}, beta={beta}, gamma={gamma}, delta={delta}"

    parser = argparse.ArgumentParser()
    parser.add_argument = MagicMock()
    argh.set_default_command(
        parser, func, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_KWONLY
    )
    _extra_kw = {"help": argh.constants.DEFAULT_ARGUMENT_TEMPLATE}
    assert parser.add_argument.mock_calls == [
        call("alpha", type=str, help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE),
        call("beta", type=str, default="N/A", nargs="?", **_extra_kw),
        call("-g", "--gamma", required=True, type=str, **_extra_kw),
        call("-d", "--delta", default="N/A", type=str, **_extra_kw),
    ]


def test_typing_hints_bool__policy_by_name_if_has_default():
    def func(
        alpha: bool, beta: bool = False, *, gamma: bool, delta: bool = False
    ) -> str:
        return f"alpha={alpha}, beta={beta}, gamma={gamma}, delta={delta}"

    parser = argparse.ArgumentParser()
    parser.add_argument = MagicMock()
    argh.set_default_command(
        parser, func, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_HAS_DEFAULT
    )
    _extra_kw = {"help": argh.constants.DEFAULT_ARGUMENT_TEMPLATE}
    assert parser.add_argument.mock_calls == [
        call("alpha", type=bool, **_extra_kw),
        call("-b", "--beta", default=False, action="store_true", **_extra_kw),
        call("gamma", type=bool, **_extra_kw),
        call("-d", "--delta", default=False, action="store_true", **_extra_kw),
    ]


def test_typing_hints_bool__policy_by_name_if_kwonly():
    def func(
        alpha: bool, beta: bool = False, *, gamma: bool, delta: bool = False
    ) -> str:
        return f"alpha={alpha}, beta={beta}, gamma={gamma}, delta={delta}"

    parser = argparse.ArgumentParser()
    parser.add_argument = MagicMock()
    argh.set_default_command(
        parser, func, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_KWONLY
    )
    _extra_kw = {"help": argh.constants.DEFAULT_ARGUMENT_TEMPLATE}
    assert parser.add_argument.mock_calls == [
        call("alpha", type=bool, **_extra_kw),
        call("beta", type=bool, default=False, nargs="?", **_extra_kw),
        call("-g", "--gamma", required=True, type=bool, **_extra_kw),
        call("-d", "--delta", default=False, action="store_true", **_extra_kw),
    ]


def test_typing_hints_literal():
    def func(
        name: Literal["Alice", "Bob"], *, greeting: Literal["Hello", "Hi"] = "Hello"
    ) -> str:
        return f"{greeting}, {name}!"

    parser = argparse.ArgumentParser()
    parser.add_argument = MagicMock()
    argh.set_default_command(
        parser, func, name_mapping_policy=NameMappingPolicy.BY_NAME_IF_KWONLY
    )
    _extra_kw = {"help": argh.constants.DEFAULT_ARGUMENT_TEMPLATE}
    assert parser.add_argument.mock_calls == [
        call("name", choices=("Alice", "Bob"), type=str, **_extra_kw),
        call(
            "-g",
            "--greeting",
            choices=("Hello", "Hi"),
            type=str,
            default="Hello",
            **_extra_kw,
        ),
    ]



#
#  Copyright © 2010—2023 Andrey Mikhaylenko and contributors
#
#  This file is part of Argh.
#
#  Argh is free software under terms of the GNU Lesser
#  General Public License version 3 (LGPLv3) as published by the Free
#  Software Foundation. See the file README.rst for copying conditions.
#
"""
Interaction
~~~~~~~~~~~
"""
from typing import Optional

__all__ = ["confirm"]


MAX_CONFIRM_ITERATIONS = 3


def confirm(
    action: str, default: Optional[bool] = None, skip: bool = False
) -> Optional[bool]:
    """
    A shortcut for typical confirmation prompt.

    :param action:

        a string describing the action, e.g. "Apply changes". A question mark
        will be appended.

    :param default:

        `bool` or `None`. Determines what happens when user hits :kbd:`Enter`
        without typing in a choice. If `True`, default choice is "yes". If
        `False`, it is "no". If `None`, the prompt keeps reappearing until user
        types in a choice (not necessarily acceptable) or until the number of
        iteration reaches the limit. Default is `None`.

    :param skip:

        `bool`; if `True`, no interactive prompt is used and default choice is
        returned (useful for batch mode). Default is `False`.

    Usage::

        def delete(key, *, silent=False):
            item = db.get(Item, args.key)
            if confirm(f"Delete {item.title}", default=True, skip=silent):
                item.delete()
                print("Item deleted.")
            else:
                print("Operation cancelled.")

    Returns `None` on `KeyboardInterrupt` event.
    """
    if skip:
        return default

    # marking the default value with a capital letter
    defaults = {
        None: ("y", "n"),
        True: ("Y", "n"),
        False: ("y", "N"),
    }
    label_yes, label_no = defaults[default]
    prompt = f"{action}? ({label_yes}/{label_no})"
    choice = None
    try:
        if default is None:
            cnt = 1
            while not choice and cnt < MAX_CONFIRM_ITERATIONS:
                choice = input(prompt)
                cnt += 1
        else:
            choice = input(prompt)
    except KeyboardInterrupt:
        return None

    if choice in ("yes", "y", "Y"):
        return True
    if choice in ("no", "n", "N"):
        return False
    if default is not None:
        return default
    return None
